/*
 *    snake - a videogame derived from Q-Basic nibbles
 *    Copyright (C) 1999 Cyberandro J.K. Starrik
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * file: ui_X.c *  purpose: handle X graphics 
 */

#include "sins.h"
#include "ui_picts.h"
#include <X11/Xlib.h> 

#undef USE_ROOT_WINDOW
#undef DEBUG_X_EVENTS
#define DEFAULTFONTCOLOR "rgb:FFFF/FFFF/0000"

#define SPRITEINMESSAGEBOX(x,y) ( (y*spriteheight)+spriteheight >= ((ydim-1)-spriteheight)-fontmaxascent && (y*spriteheight) <= ((ydim-1)-spriteheight)+fontmaxdescent )

/* public functions */
int ui_init ();
int ui_initarena (ARENA *);
void ui_finish (void);
void ui_drawarena (ARENA *);
void ui_drawmessage (char *);
char *ui_prompt (char *);
int ui_getkey (void);
int CMD_loadpictset (char *line);
int CMD_fontcolor (char *line);
int CMD_loadfont (char *pattern);

/* private functions */
static void init_background(ARENA *);
static void draw_pict (Drawable d, int sx, int sy, int pictnum);
static int *load_pictset (char *filename);
#ifndef USE_ROOT_WINDOW
static Window create_arenawin (Display *dpy, int width, int height);
#endif
static void create_pictset_pixmap (void);
static void enable_backingstore(Display *d, Window w);
static int send_pictset (void);
static int copy_pict (int x, int y, int pictnum);
static void redraw_arenarect(int rx, int ry, int rw, int rh);
static void create_graphic_contexts (void);
static void init_colors (void);
static int load_font (char *);
static void redrawmessage (void);

/* private data */
static char *arenamap; /* picture indexes */
static int spriteheight;
static int spritewidth;
static int *pictset;
static int pictsetsent = 0;
static int coloridx[16]; /* index of default 16 colors */

static Display *dpy;
static Screen *scr;
static Window arenawin;
static Pixmap pictset_pixmap;
static GC arenagc; 
static int ydim, xdim;
static Font font = -1;
static int fontcolor = 0;
static int fontmaxascent;
static int fontmaxdescent;
static int fontmaxlbearing;
static int fontmaxrbearing;
static char *lastmessage;


/* public data */
int screen_initialized = 0;

/* Initialize UI module */
int
ui_init ()
{
  if (!register_command ("loadpictset", &CMD_loadpictset,
			 "loadpictset [<fname>]")) return 0;

  if (!register_command ("fontcolor", &CMD_fontcolor,
			 "fontcolor [<colordef>]")) return 0;

  if (!register_command ("loadfont", &CMD_loadfont,
			 "loadfont [<pattern>]")) return 0;

  pictset = defpictset;

  return 1;
}

/* initialize the graphics */
int
ui_initarena (ARENA *arena)
{

  /* Connect to the X server */
  if ( ! (dpy=XOpenDisplay(NULL)) ) 
  {
    message ("Can't open display: %s", getenv("DISPLAY"));
    finish();
  }

  scr = DefaultScreenOfDisplay(dpy);

  /* default dimensions */
#ifdef USE_ROOT_WINDOW
  xdim = scr->width;
  ydim = scr->height;
#else
  xdim = (scr->width) * 0.7;
  ydim = (scr->height) * 0.7;
#endif

  /* Set window dimensions based on arena and sprites dimensions */
  if ( !spritewidth ) spritewidth=8;
  if ( !spriteheight ) spriteheight=8;
  if ( arena->lines )
  {
    /* set window height based on arena and sprite heights */
    ydim = spriteheight * ( arena->lines );
  }
  else
  {
    /* set arena height based on window and sprite heights */
    arena->lines = ydim/spriteheight;
  }
  if ( arena->cols )
  {
    /* set window width based on arena and sprite widths */
    xdim = spritewidth * ( arena->cols );
  }
  else
  {
    /* set arena width based on window and sprite widths */
    arena->cols = xdim/spritewidth;
  }

  create_pictset_pixmap();

#ifdef USE_ROOT_WINDOW
  arenawin = DefaultRootWindow(dpy);
  XSelectInput(dpy, arenawin, ExposureMask);
#else
  arenawin = create_arenawin(dpy, xdim, ydim);
#endif


  /* We should not rely on this */
  enable_backingstore(dpy, arenawin);

  create_graphic_contexts();

  /* load the first font found */
  if ( ! load_font("*") )
  {
    finish();
  }

  /* Set colors index */
  init_colors();

  send_pictset();

  init_background(arena);

  /*
   * cache server's KeySym now ... just to avoid later delay.
   * there should be a better way, but i'm an hacker after all ...
   */
  XKeycodeToKeysym(dpy, 10, 0);

  XFlush(dpy);

  /* initialize the arena maps (used for optimization purposes) */
  arenamap = (char *)safe_malloc( arena->lines*arena->cols*sizeof(int),
    "ui_X: ui_initarena: arenamap");
  memset(arenamap, -1, arena->lines * arena->cols);

  /* Set screen initialized flag */
  screen_initialized = 1;

  return 1;

}

/* X version of message drawing function */
void
ui_drawmessage (char *str)
{

  lastmessage = str;

  if ( font == -1 )
  {
    if ( strlen(str) ) fprintf(stderr, "%s\n", str);
    return;
  }

  XSetForeground(dpy, arenagc, fontcolor);

  redraw_arenarect(
    (0+spritewidth)-fontmaxlbearing, (ydim-1-spriteheight)-fontmaxascent,
    (xdim-1), fontmaxascent+fontmaxdescent);

  XDrawString(dpy, arenawin, arenagc, 0+spritewidth,
    (ydim-1)-spriteheight, str, strlen(str));
  
  return;
}

static void
redrawmessage ()
{

  if ( font == -1 ) return;

  if ( ! lastmessage || ! *lastmessage ) return;

  XSetForeground(dpy, arenagc, fontcolor);

  redraw_arenarect(
    (0+spritewidth)-fontmaxlbearing, (ydim-1-spriteheight)-fontmaxascent,
    (xdim-1), fontmaxascent+fontmaxdescent);

  XDrawString(dpy, arenawin, arenagc, 0+spritewidth,
    (ydim-1)-spriteheight, lastmessage, strlen(lastmessage));

  return;
}

void
ui_drawarena (ARENA *arena)
{
  int y, x;
  sprite *sp;
  int pictidx;
  int should_redraw = 0;
  int should_redraw_message = 0;
  XEvent event;

#ifndef USE_ROOT_WINDOW
  while ( XCheckWindowEvent(dpy, arenawin, StructureNotifyMask|ExposureMask, &event) )
#else
  while ( XCheckWindowEvent(dpy, arenawin, ExposureMask, &event) )
#endif
  {
    switch(event.type)
    {

#ifndef USE_ROOT_WINDOW
      case (ConfigureNotify):

#ifdef DEBUG_X_EVENTS
        message("Event: ConfigureNotify (width=%i, height=%i)",
          event.xconfigure.width, event.xconfigure.height);
#endif
        if (event.xconfigure.width != xdim || event.xconfigure.height != ydim)
        {
          xdim = event.xconfigure.width;
          ydim = event.xconfigure.height;
          message("window resize");
          should_redraw = 1;
        }
        break;
#endif

      case (Expose):
#ifdef DEBUG_X_EVENTS
        message("Event: Expose (x=%i, y=%i, w=%i, h=%i)", event.xexpose.x,
          event.xexpose.y, event.xexpose.width, event.xexpose.height);
#endif
        redraw_arenarect(event.xexpose.x, event.xexpose.y,
          event.xexpose.width, event.xexpose.height);
        break;
        
      default:
#ifdef DEBUG_X_EVENTS
        message("Event: %i", event.type);
#endif
        break;
    }
  }

  /* Send pictset to the display if needed */
  if ( ! pictsetsent )
  {
    send_pictset();
    should_redraw = 1;
  }

  for (y = 0; y < arena->lines; y++)
  {
    for (x = 0; x < arena->cols; x++)
    {

      sp = GETSPRITE(arena,y,x);

      if (sp) switch (sp->type)
      {
        case SNAKE:
          pictidx = PICT_SNAKE0+((snakebody *)(sp))->self->playernum;
          break;  
        case EXSNAKE:
          pictidx = PICT_EXSNAKE;
          break; 
        case FRUIT:
          pictidx = PICT_FRUIT0+(((fruit *)(sp))->value)-48;
          break;
        case WALL:
          pictidx = PICT_WALL;
          break;
        default:
          pictidx = PICT_UNKNOWN;
          break;
      }
      else pictidx = PICT_EMPTY;

      if ( *(arenamap+y*arena->cols+x) == pictidx && !should_redraw ) continue;
      *(arenamap+y*arena->cols+x) = pictidx;

      copy_pict(x, y, pictidx);

      /* find wheter message should be redrawn */
      if ( SPRITEINMESSAGEBOX(x,y) ) should_redraw_message = 1;
    }
  }

  if ( should_redraw || should_redraw_message )
  {
    redrawmessage();
  }

  XFlush(dpy);
}


void
ui_finish ()
{
  if ( screen_initialized )
  {
    XCloseDisplay(dpy);
  }
}


/*
 * Get next key in input buffer
 *
 * OK. We should inspect keyboard status here.
 * It's not that easy. The calling function expect ONE and only ONE char
 * to be returned at a time. It's used to ``consume'' the input buffer
 * before proceding. We need to feed it with all the key pressed up to
 * now I guess... How ?!
 *
 */
int
ui_getkey ()
{
  int retkey = -1;

#ifndef USE_ROOT_WINDOW
  int key = -1;
  XEvent event;
  static int shift = 0;

  while ( XCheckWindowEvent(dpy, arenawin, KeyPressMask|KeyReleaseMask, &event) )
  {

#ifdef DEBUG_X_EVENTS
    switch (event.type)
    {
      case KeyPress:
        message("Event: KeyPress (keycode=%i)", event.xkey.keycode);
        break;
      case KeyRelease:
        message("Event: KeyRelease (keycode=%i)", event.xkey.keycode);
        break;
      default:
        message("Event: %i", event.type);
        break;
    }
#endif

    key = XKeycodeToKeysym(dpy, event.xkey.keycode, shift);
    /* message("ui_getkey: keysym %i / keycode %i", key, event.xkey.keycode); */

    /* handle a shift */
    if ( key == 65505 || key == 65506 )
    {
      shift = event.type == KeyRelease ? 0 : 1;
      continue;
    }

    if ( event.type == KeyPress )
    {

      /* ENTER KEY */
      if ( key == 65293 ) retkey = '\n';
      else if ( key == 65288 ) retkey = 127;

      /* I found this to be working for HIGH-valued keysyms 
       * My nice game segfaults if this function returns numbers
       * higher then 255 (who knows why ? -- me, a char type is expected)
       */
      else if ( key > 256 ) retkey = key % 256;
      else retkey = key;

      break;
    }

  }
#endif

  return retkey;
}

/*
 * prompt the user for a string
 * the line returnered is allocated with malloc(3)
 * so the caller must free it when finished with it. 
 */
char *
ui_prompt (char *prompt_string)
{
  int key;
  int len = 0;
  char *cmdline;

  cmdline = safe_malloc (MAXCMDLINELEN, "prompt input buffer");
  memset(cmdline, '\0', MAXCMDLINELEN);

  message("%s: %s_", prompt_string, cmdline);
  ui_drawarena(&arena);

  while (1)
  {

    key=ui_getkey();
    if ( key == -1 ) continue;

    /* max command line lenght reached */
    if ( len > MAXCMDLINELEN ) break;

    if ( key == '\n' ) break;

    /* handle backspace */
    if ( key == 127 )
    {
        if (len) cmdline[--len] = '\0';
    }

    else cmdline[len++] = key;
    message("%s: %s_", prompt_string, cmdline);

    ui_drawarena(&arena);

  }
  
  ui_drawmessage("");
  return cmdline;
}

int
CMD_loadpictset (char *line)
{
  char fname[MAXPATHLEN];
  char *fpath;
  int ret = 0;
  int *newset;

  if ( line ) ret = sscanf (line, "%s", fname);
  if ( ret != 1 )
  {
    /* set default pictset (should instead complain about wrong num of args?) */
    pictset = defpictset;
    message("default pictset loaded");
    pictsetsent = 0;
    return 1;
  }

  if ( fname[0] != '/' )
  {
    if ( (fpath=find_file(fname)) == NULL )
    {
            message ("Can't find file %s in path %s", fname, path);
            return 0;
    }
  }
  else
  {
    fpath = fname;
  }

  newset = load_pictset(fpath);
  if ( !newset )
  {
    message("Can't load pictset from %s", fpath);
    return 0;
  }

  if ( pictset != defpictset ) free(pictset);
  pictset = newset;

  pictsetsent = 0;

  message("pictset %s loaded", fname);
  return 1;
}

int
CMD_fontcolor (char *colordef)
{
  XColor color;
  Colormap cm;

  if ( ! screen_initialized )
  {
    message("fontcolor: No connection to the X-server yet..");
    return 0;
  }

  if ( ! colordef ) colordef = DEFAULTFONTCOLOR;
 
  cm = DefaultColormap(dpy, DefaultScreen(dpy));
  if ( ! XParseColor(dpy, cm, colordef, &color) )
  {
      message("color %s not found", DEFAULTFONTCOLOR);
      return 0;
  }
  XAllocColor(dpy, cm, &color);
  fontcolor = color.pixel;
  message("Font color changed to %s", colordef);
  return 1;
}

int
CMD_loadfont (char *pattern)
{
  if ( ! screen_initialized )
  {
    message("loadfont: No connection to the X-server yet..");
    return 0;
  }

  if ( ! pattern ) pattern = "*";
  return load_font(pattern);
}


/************ PRIVATE FUNCTIONS **********************/


static void
enable_backingstore (Display *d, Window w)
{
  XSetWindowAttributes wa;
  wa.backing_store = Always;
  XChangeWindowAttributes(d, w, CWBackingStore, &wa);
}

#ifndef USE_ROOT_WINDOW
static Window
create_arenawin (Display *dpy, int width, int height)
{
  Window w;

  w = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0,
    width, height, 0, 0, 0);

  message("Created arena window %i", w);

  /* We want to get MapNotify events */
  XSelectInput(dpy, w, StructureNotifyMask);

  /* "Map" the window (that is, make it appear on the screen) */
  XMapWindow(dpy, w);

  for(;;)
  {
    XEvent e;
    XNextEvent(dpy, &e);
    if (e.type == MapNotify) break;
  }

  /* Select events */
  /* XSelectInput(dpy, w, KeyPressMask|FocusChangeMask|StructureNotifyMask); */
  XSelectInput(dpy, w, KeyPressMask|KeyReleaseMask|StructureNotifyMask|ExposureMask);
  message("Selected events on window %i", w);

  /* Give window a name */
  XStoreName (dpy, w, "X-Sins");

  return w;
}
#endif


static void 
create_pictset_pixmap (void)
{

  pictset_pixmap = XCreatePixmap (dpy, DefaultRootWindow(dpy), spritewidth,
    spriteheight*PICTSETLEN, DefaultDepthOfScreen(scr));

  message("Created pictset pixmap %i", pictset_pixmap);

}

static void
init_background(ARENA *arena)
{
  int y,x;

  /* initialize background */
  for (y = 0; y < arena->lines; y++)
  {
    for (x = 0; x < arena->cols; x++)
    {
        copy_pict(x, y, PICT_EMPTY);
    }
  }
}

static void
draw_pict (Drawable w, int sx, int sy, int pictnum)
{
  int x,y,i,j;
  int *pict;
  int color;

  sx=sx*spritewidth;
  sy=sy*spriteheight;
  pict=pictset+pictnum*spritewidth*spriteheight;

  for (x=sx, i=0; i<spritewidth; i++, x++)
  {
    for (y=sy, j=0; j<spriteheight; y++, j++)
    {
      color = *(pict+(j*spritewidth)+i);
      /* XSetForeground( dpy, arenagc, color ); */
      XSetForeground( dpy, arenagc, coloridx[color] );
      XDrawPoint( dpy, w, arenagc, x, y );
    }
  }
}

/*
 * Copy a picture from the pictset pixmap to the specified
 * location on the arenawin
 */
static int
copy_pict (int x, int y, int pictnum)
{
  int sx, sy, dx, dy;

  sx = 0;
  sy = pictnum*spriteheight;
  dx=x*spritewidth;
  dy=y*spriteheight;

  /*
   * This generates NoExpose and GraphicsExpose event 
   * unless the GC has the graphics_exposures set to False
   */
  XCopyArea(dpy,pictset_pixmap,arenawin,arenagc,sx,sy,spritewidth,spriteheight,dx,dy);

  return 1;
}



static int *
load_pictset (char *filename)
{
  FILE *stream;
  int idx = 0;
  int quoted = 0;
  int ch;
  int *set;
  int w, h;

  stream = fopen(filename, "r");
  if ( !stream ) return NULL;

  /* check sprite size */
  if ( fscanf(stream, "#%ix%i\n", &w, &h) != 2 )
  {
    message("Invalid pictset format (missing '#WxH' on line 1)");
    return NULL;
  }

  if ( screen_initialized && (w != spritewidth || h != spriteheight) )
  {
    message("Can't change sprite size while running");
    return NULL;
  }

  spritewidth = w;
  spriteheight = h;
  set = safe_malloc(w*h*PICTSETLEN*sizeof(int), "New pictset");

  while (!feof(stream))
  {
    if ( idx >= w*h*PICTSETLEN ) break;
    ch=getc(stream);
    if (ch) switch(ch)
    {
      case '\n':
      case '\r':
        quoted=0;
        break;
      case ' ':
      case '\t':
        break;
      case '#':
        quoted=1;
        break;
      default:
        if ( !quoted )
        {
          *(set+idx) = ch-97;
          idx++;
        }
        break;
    }
  }
  fclose(stream);
  return set;
}

static int
send_pictset()
{
  int n;

  for (n=0; n<PICTSETLEN; n++)
  {
    draw_pict(pictset_pixmap, 0, n, n);
  }

  message("Pictset sent");
  pictsetsent = 1;

  return 1;
}

static void
redraw_arenarect(int rx, int ry, int rw, int rh)
{
  int ax, ay, aw, ah;
  int y, x;
  int pictidx;

  ax = rx/spritewidth;
  ay = ry/spriteheight;
  aw = rw/spritewidth;
  if ( rw%spritewidth ) aw++;
  ah = rh/spriteheight;
  if ( rh%spriteheight ) ah++;

  for (y = ay; y-ay <= ah && y < ARENA_HEIGHT; y++)
  {
    for (x = ax; x-ax <= aw && x < ARENA_WIDTH; x++)
    {
      pictidx = *(arenamap+y*arena.cols+x);
      copy_pict(x, y, pictidx);
    }
  }

  XFlush(dpy);
}


static void
create_graphic_contexts()
{
  XGCValues gcval;

  gcval.graphics_exposures = False;


  /*
   * Create the Graphic Context
   * Set the graphics-exposures to False to avoid
   * NoExpose and GraphicsExpose Event on XCopyArea requests
   *
   * Anyway... NoExpose events suggest there are some problems with XCopyArea
   * ( unreachable source coordinates ? )
   */ 
  arenagc = XCreateGC(dpy, arenawin, GCGraphicsExposures, &gcval);

}

static void
init_colors()
{
  Colormap cm;
  int i;
  XColor color;
  char *colornames[16] = {
    "rgb:0000/0000/0000", /* black */
    "rgb:0000/0000/AAAA", /* blue */
    "rgb:0000/8888/0000", /* green */
    "rgb:0000/8888/8888", /* cyan */
    "rgb:8888/0000/0000", /* red */
    "rgb:8888/8888/0000", /* magenta */
    "rgb:8888/0000/8888", /* brown */
    "rgb:8888/8888/8888", /* grey */

    "rgb:4444/4444/4444", /* light black */
    "rgb:0000/0000/FFFF", /* light blue */
    "rgb:0000/FFFF/0000", /* light green */
    "rgb:0000/FFFF/FFFF", /* light cyan */
    "rgb:FFFF/0000/0000", /* light red */
    "rgb:FFFF/FFFF/0000", /* light magenta */
    "rgb:FFFF/0000/FFFF", /* light brown */
    "rgb:FFFF/FFFF/FFFF", /* light grey */
  };

  cm = DefaultColormap(dpy, DefaultScreen(dpy));
  /* message("Default colormap = %i", cm); */

  for (i=0; i<16; i++)
  {
    if ( ! XParseColor(dpy, cm, colornames[i], &color) )
    {
      message("color %s not found", colornames[i]);
    }
    XAllocColor(dpy, cm, &color);
    coloridx[i] = color.pixel;
  }

  if ( ! XParseColor(dpy, cm, DEFAULTFONTCOLOR, &color) )
  {
      message("color %s not found", DEFAULTFONTCOLOR);
      finish();
  }
  XAllocColor(dpy, cm, &color);
  fontcolor = color.pixel;
  message("font color is %i", fontcolor);

}


/*
 * Load font
 */
static int
load_font (char *pattern)
{
  XFontStruct *fontinfo;
  char **fontnames;
  int foundfonts = 0;

  if ( ! pattern ) pattern = "*";

  if ( ! (fontnames=XListFontsWithInfo(dpy, pattern, 1, &foundfonts, &fontinfo)) || ! foundfonts )
  {
    message("No fonts matching pattern %s", pattern);
    return 0;
  }

  /* Unload old font if any */
  if ( font != -1 ) XUnloadFont(dpy, font);

  /* Load the font */
  font = XLoadFont(dpy, fontnames[0]);

  /* Write down font dimensions */
  fontmaxascent = fontinfo->max_bounds.ascent;
  fontmaxdescent = fontinfo->max_bounds.descent;
  fontmaxlbearing = fontinfo->max_bounds.rbearing;
  fontmaxrbearing = fontinfo->max_bounds.lbearing;

  XFreeFontInfo (fontnames, fontinfo, foundfonts);

  XSetFont(dpy, arenagc, font);
  message("Font %i set in arenagc", font);

  return 1;
}

